import { ApiProperty } from '@nestjs/swagger';
import { Transform } from 'class-transformer';
import { IsDate, IsEnum, IsInt, IsNotEmpty, IsNumber, IsPositive, IsString, IsUUID, ValidateIf } from 'class-validator';
import moment from 'moment';
import { i18nValidationMessage as i18n } from 'nestjs-i18n';
import { AllowanceFrequency, AllowanceType } from '~/allowance/enums';
export class CreateAllowanceRequestDto {
  @ApiProperty({ example: 'Allowance name' })
  @IsString({ message: i18n('validation.IsString', { path: 'general', property: 'allowance.name' }) })
  @IsNotEmpty({ message: i18n('validation.IsNotEmpty', { path: 'general', property: 'allowance.name' }) })
  name!: string;

  @ApiProperty({ example: 100 })
  @IsNumber({}, { message: i18n('validation.IsNumber', { path: 'general', property: 'allowance.amount' }) })
  @IsPositive({ message: i18n('validation.IsPositive', { path: 'general', property: 'allowance.amount' }) })
  amount!: number;

  @ApiProperty({ example: AllowanceFrequency.WEEKLY })
  @IsEnum(AllowanceFrequency, {
    message: i18n('validation.IsEnum', { path: 'general', property: 'allowance.frequency' }),
  })
  frequency!: AllowanceFrequency;

  @ApiProperty({ example: AllowanceType.BY_END_DATE })
  @IsEnum(AllowanceType, { message: i18n('validation.IsEnum', { path: 'general', property: 'allowance.type' }) })
  type!: AllowanceType;

  @ApiProperty({ example: new Date() })
  @IsDate({ message: i18n('validation.IsDate', { path: 'general', property: 'allowance.startDate' }) })
  @Transform(({ value }) => moment(value).startOf('day').toDate())
  startDate!: Date;

  @ApiProperty({ example: new Date() })
  @IsDate({ message: i18n('validation.IsDate', { path: 'general', property: 'allowance.endDate' }) })
  @Transform(({ value }) => moment(value).endOf('day').toDate())
  @ValidateIf((o) => o.type === AllowanceType.BY_END_DATE)
  endDate?: Date;

  @ApiProperty({ example: 10 })
  @IsNumber(
    {},
    { message: i18n('validation.IsNumber', { path: 'general', property: 'allowance.numberOfTransactions' }) },
  )
  @IsInt({ message: i18n('validation.IsInt', { path: 'general', property: 'allowance.amount' }) })
  @IsPositive({ message: i18n('validation.IsPositive', { path: 'general', property: 'allowance.amount' }) })
  @ValidateIf((o) => o.type === AllowanceType.BY_COUNT)
  numberOfTransactions?: number;

  @ApiProperty({ example: 'e7b1b3b4-4b3b-4b3b-4b3b-4b3b4b3b4b3b' })
  @IsUUID('4', { message: i18n('validation.IsUUID', { path: 'general', property: 'allowance.juniorId' }) })
  juniorId!: string;
}
