import { Injectable, Logger } from '@nestjs/common';
import { Cron, CronExpression } from '@nestjs/schedule';
import moment from 'moment';
import { AllowancesService } from '~/allowance/services';
import { CacheService } from '~/common/modules/cache/services';
import { BaseCronService } from '../services/base-cron.service';
const TEN = 10;
const SIXTY = 60;
const THOUSAND = 1000;
const TEN_MINUTES = TEN * SIXTY * THOUSAND;
const CHUNK_SIZE = 50;

@Injectable()
export class AllowanceTask extends BaseCronService {
  private readonly logger = new Logger(AllowanceTask.name);
  private readonly cronLockKey = `${AllowanceTask.name}-lock`;
  private readonly cronLockTtl = TEN_MINUTES;
  constructor(cacheService: CacheService, private readonly allowanceService: AllowancesService) {
    super(cacheService);
  }

  @Cron(CronExpression.EVERY_DAY_AT_3AM)
  async handleCron() {
    try {
      const isLockAcquired = await this.acquireLock(this.cronLockKey, this.cronLockTtl);
      if (!isLockAcquired) {
        this.logger.log('Lock already acquired. Skipping cron job.');
        return;
      }
      this.logger.log('Processing cron job');
      await this.processJob();
    } catch (error) {
      this.logger.error('Error processing cron job', error);
    } finally {
      this.logger.log('Releasing lock');
      await this.releaseLock(this.cronLockKey);
    }
  }

  private async processJob() {
    const today = moment().startOf('day');
    const allowancesChunks = await this.allowanceService.findAllowancesChunks(CHUNK_SIZE);
    for await (const allowances of allowancesChunks) {
      for (const allowance of allowances) {
        this.logger.log(`Processing allowance ${allowance.id} with next payment date ${allowance.nextPaymentDate}`);
        // if today is the same as allowance payment date
        if (moment(allowance.nextPaymentDate).startOf('day').isSame(today)) {
          this.logger.log(`Today is the payment date for allowance ${allowance.id}`);
          //@TODO: Implement payment logic
        }
      }
    }
  }
}
