import { Injectable, Logger } from '@nestjs/common';
import { Cron, CronExpression } from '@nestjs/schedule';
import moment from 'moment';
import { CacheService } from '~/common/modules/cache/services';
import { MoneyRequestsService } from '~/money-request/services';
import { BaseCronService } from '../services';
const TEN = 10;
const SIXTY = 60;
const THOUSAND = 1000;
const TEN_MINUTES = TEN * SIXTY * THOUSAND;
const CHUNK_SIZE = 50;
@Injectable()
export class MoneyRequestTask extends BaseCronService {
  private readonly cronLockKey = `${MoneyRequestTask.name}-lock`;
  private readonly cronLockTtl = TEN_MINUTES;
  private readonly logger = new Logger(MoneyRequestTask.name);
  constructor(cacheService: CacheService, private readonly moneyRequestService: MoneyRequestsService) {
    super(cacheService);
  }

  @Cron(CronExpression.EVERY_DAY_AT_3AM)
  async handleCron() {
    try {
      const isLockAcquired = await this.acquireLock(this.cronLockKey, this.cronLockTtl);
      if (!isLockAcquired) {
        this.logger.log('Lock already acquired. Skipping cron job for MoneyRequestTask.');
        return;
      }
      this.logger.log('Processing cron job for MoneyRequestTask');
      await this.processJob();
    } catch (error) {
      this.logger.error('Error processing MoneyRequestTask cron job', error);
    } finally {
      this.logger.log('Releasing lock for MoneyRequestTask');
      await this.releaseLock(this.cronLockKey);
    }
  }

  private async processJob() {
    const today = moment().startOf('day');
    const moneyRequestsChunks = await this.moneyRequestService.findMoneyRequestsChunks(CHUNK_SIZE);
    for await (const moneyRequests of moneyRequestsChunks) {
      for (const moneyRequest of moneyRequests) {
        this.logger.log(
          `Processing money request ${moneyRequest.id} with next payment date ${moneyRequest.nextPaymentDate}`,
        );
        // if today is the same as money request payment date
        if (moment(moneyRequest.nextPaymentDate).startOf('day').isSame(today)) {
          this.logger.log(`Today is the payment date for money request ${moneyRequest.id}`);
        }
      }
    }
  }
}
