import { BadRequestException, Injectable, Logger } from '@nestjs/common';
import { Roles } from '~/auth/enums';
import { IJwtPayload } from '~/auth/interfaces';
import { OciService } from '~/document/services';
import { JuniorService } from '~/junior/services';
import { CreateGiftRequestDto, GiftFiltersRequestDto, GiftReplyRequestDto } from '../dtos/request';
import { Gift } from '../entities';
import { GiftStatus } from '../enums';
import { GiftsRepository } from '../repositories';

@Injectable()
export class GiftsService {
  private readonly logger = new Logger(GiftsService.name);
  constructor(
    private readonly juniorService: JuniorService,
    private readonly giftsRepository: GiftsRepository,
    private readonly ociService: OciService,
  ) {}

  async createGift(guardianId: string, body: CreateGiftRequestDto) {
    this.logger.log(`Creating gift for junior ${body.recipientId} by guardian ${guardianId}`);
    const doesJuniorBelongToGuardian = await this.juniorService.doesJuniorBelongToGuardian(
      guardianId,
      body.recipientId,
    );

    if (!doesJuniorBelongToGuardian) {
      this.logger.error(`Junior ${body.recipientId} does not belong to guardian ${guardianId}`);
      throw new BadRequestException('JUNIOR.DOES_NOT_BELONG_TO_GUARDIAN');
    }

    const gift = await this.giftsRepository.create(guardianId, body);
    this.logger.log(`Gift ${gift.id} created successfully`);
    return this.findUserGiftById({ sub: guardianId, roles: [Roles.GUARDIAN] }, gift.id);
  }

  async findUserGiftById(user: IJwtPayload, giftId: string) {
    this.logger.log(`Finding gift ${giftId} for user ${user.sub} with roles ${user.roles}`);
    const gift = user.roles.includes(Roles.GUARDIAN)
      ? await this.giftsRepository.findGuardianGiftById(user.sub, giftId)
      : await this.giftsRepository.findJuniorGiftById(user.sub, giftId);

    if (!gift) {
      this.logger.error(`Gift ${giftId} not found`);
      throw new BadRequestException('GIFT.NOT_FOUND');
    }

    await this.prepareGiftImages([gift]);

    this.logger.log(`Gift ${giftId} found successfully`);

    return gift;
  }

  async redeemGift(juniorId: string, giftId: string) {
    this.logger.log(`Redeeming gift ${giftId} for junior ${juniorId}`);
    const gift = await this.giftsRepository.findJuniorGiftById(juniorId, giftId);

    if (!gift) {
      this.logger.error(`Gift ${giftId} not found`);
      throw new BadRequestException('GIFT.NOT_FOUND');
    }

    if (gift.status === GiftStatus.REDEEMED) {
      this.logger.error(`Gift ${giftId} already redeemed`);
      throw new BadRequestException('GIFT.ALREADY_REDEEMED');
    }

    return this.giftsRepository.redeemGift(gift);
  }

  async UndoGiftRedemption(juniorId: string, giftId: string) {
    this.logger.log(`Undoing gift redemption for junior ${juniorId} and gift ${giftId}`);
    const gift = await this.giftsRepository.findJuniorGiftById(juniorId, giftId);

    if (!gift) {
      this.logger.error(`Gift ${giftId} not found`);
      throw new BadRequestException('GIFT.NOT_FOUND');
    }

    if (gift.status === GiftStatus.AVAILABLE) {
      this.logger.error(`Gift ${giftId} is not redeemed yet`);
      throw new BadRequestException('GIFT.NOT_REDEEMED');
    }

    return this.giftsRepository.UndoGiftRedemption(gift);
  }

  async replyToGift(juniorId: string, giftId: string, body: GiftReplyRequestDto) {
    this.logger.log(`Replying to gift ${giftId} for junior ${juniorId}`);
    const gift = await this.giftsRepository.findJuniorGiftById(juniorId, giftId);

    if (!gift) {
      this.logger.error(`Gift ${giftId} not found`);
      throw new BadRequestException('GIFT.NOT_FOUND');
    }

    if (gift.reply) {
      this.logger.error(`Gift ${giftId} already replied`);
      throw new BadRequestException('GIFT.ALREADY_REPLIED');
    }

    return this.giftsRepository.replyToGift(gift, body);
  }

  findGifts(user: IJwtPayload, filters: GiftFiltersRequestDto) {
    this.logger.log(`Finding gifts for user ${user.sub} with roles ${user.roles}`);
    return this.giftsRepository.findGifts(user, filters);
  }

  private async prepareGiftImages(gifts: Gift[]) {
    this.logger.log('Preparing gifts image');
    await Promise.all(
      gifts.map(async (gift) => {
        gift.image.url = await this.ociService.generatePreSignedUrl(gift.image);
      }),
    );
  }
}
