import { BadRequestException, Injectable, Logger } from '@nestjs/common';
import { Transactional } from 'typeorm-transactional';
import { Roles } from '~/auth/enums';
import { PageOptionsRequestDto } from '~/core/dtos';
import { CustomerNotificationSettings } from '~/customer/entities/customer-notification-settings.entity';
import { CustomerService } from '~/customer/services';
import { UserService } from '~/user/services';
import { CreateJuniorRequestDto, SetThemeRequestDto } from '../dtos/request';
import { Junior } from '../entities';
import { JuniorRepository } from '../repositories';
import { JuniorTokenService } from './junior-token.service';

@Injectable()
export class JuniorService {
  private readonly logger = new Logger(JuniorService.name);
  constructor(
    private readonly juniorRepository: JuniorRepository,
    private readonly juniorTokenService: JuniorTokenService,
    private readonly userService: UserService,
    private readonly customerService: CustomerService,
  ) {}

  @Transactional()
  async createJuniors(body: CreateJuniorRequestDto, guardianId: string) {
    this.logger.log(`Creating junior for guardian ${guardianId}`);
    const existingUser = await this.userService.findUser([{ email: body.email }, { phoneNumber: body.phoneNumber }]);

    if (existingUser) {
      this.logger.error(`User with email ${body.email} or phone number ${body.phoneNumber} already exists`);
      throw new BadRequestException('USER.ALREADY_EXISTS');
    }

    const user = await this.userService.createUser({
      email: body.email,
      countryCode: body.countryCode,
      phoneNumber: body.phoneNumber,
      roles: [Roles.JUNIOR],
    });

    await this.customerService.createCustomer(
      {
        firstName: body.firstName,
        lastName: body.lastName,
        dateOfBirth: body.dateOfBirth,
        junior: Junior.create({
          id: user.id,
          guardianId,
          relationship: body.relationship,
          civilIdFrontId: body.civilIdFrontId,
          civilIdBackId: body.civilIdBackId,
        }),
        notificationSettings: new CustomerNotificationSettings(),
      },
      user,
    );

    this.logger.log(`Junior ${user.id} created successfully`);

    return this.juniorTokenService.generateToken(user.id);
  }

  async findJuniorById(juniorId: string, withGuardianRelation = false, guardianId?: string) {
    this.logger.log(`Finding junior ${juniorId}`);
    const junior = await this.juniorRepository.findJuniorById(juniorId, withGuardianRelation, guardianId);

    if (!junior) {
      this.logger.error(`Junior ${juniorId} not found`);
      throw new BadRequestException('JUNIOR.NOT_FOUND');
    }

    this.logger.log(`Junior ${juniorId} found successfully`);
    return junior;
  }

  @Transactional()
  async setTheme(body: SetThemeRequestDto, juniorId: string) {
    this.logger.log(`Setting theme for junior ${juniorId}`);
    const junior = await this.findJuniorById(juniorId);
    if (junior.theme) {
      this.logger.log(`Removing existing theme for junior ${juniorId}`);
      await this.juniorRepository.removeTheme(junior.theme);
    }

    await this.juniorRepository.setTheme(body, junior);
    this.logger.log(`Theme set for junior ${juniorId}`);
    return this.juniorRepository.findThemeForJunior(juniorId);
  }

  findJuniorsByGuardianId(guardianId: string, pageOptions: PageOptionsRequestDto) {
    this.logger.log(`Finding juniors for guardian ${guardianId}`);
    return this.juniorRepository.findJuniorsByGuardianId(guardianId, pageOptions);
  }

  async validateToken(token: string) {
    this.logger.log(`Validating token ${token}`);
    const juniorId = await this.juniorTokenService.validateToken(token);
    return this.findJuniorById(juniorId, true);
  }

  generateToken(juniorId: string) {
    this.logger.log(`Generating token for junior ${juniorId}`);
    return this.juniorTokenService.generateToken(juniorId);
  }

  async doesJuniorBelongToGuardian(guardianId: string, juniorId: string) {
    this.logger.log(`Checking if junior ${juniorId} belongs to guardian ${guardianId}`);
    const junior = await this.findJuniorById(juniorId, false, guardianId);

    return !!junior;
  }
}
