import moment from 'moment';
import {
  Column,
  CreateDateColumn,
  Entity,
  JoinColumn,
  ManyToOne,
  PrimaryGeneratedColumn,
  UpdateDateColumn,
} from 'typeorm';
import { Guardian } from '~/guardian/entities/guradian.entity';
import { Junior } from '~/junior/entities';
import { MoneyRequestFrequency, MoneyRequestStatus } from '../enums';

@Entity('money_requests')
export class MoneyRequest {
  @PrimaryGeneratedColumn('uuid')
  id!: string;

  @Column({ type: 'decimal', precision: 10, scale: 3, name: 'requested_amount' })
  requestedAmount!: number;

  @Column({ type: 'varchar', name: 'message' })
  message!: string;

  @Column({ type: 'varchar', name: 'frequency', default: MoneyRequestFrequency.ONE_TIME })
  frequency!: MoneyRequestFrequency;

  @Column({ type: 'varchar', name: 'status', default: MoneyRequestStatus.PENDING })
  status!: MoneyRequestStatus;

  @Column({ type: 'timestamp with time zone', name: 'reviewed_at', nullable: true })
  reviewedAt!: Date;

  @Column({ type: 'timestamp with time zone', name: 'start_date', nullable: true })
  startDate!: Date | null;

  @Column({ type: 'timestamp with time zone', name: 'end_date', nullable: true })
  endDate!: Date | null;

  @Column({ type: 'uuid', name: 'requester_id' })
  requesterId!: string;

  @Column({ type: 'uuid', name: 'reviewer_id' })
  reviewerId!: string;

  @ManyToOne(() => Junior, (junior) => junior.moneyRequests)
  @JoinColumn({ name: 'requester_id' })
  requester!: Junior;

  @ManyToOne(() => Guardian, (guardian) => guardian.moneyRequests)
  @JoinColumn({ name: 'reviewer_id' })
  reviewer!: Guardian;

  @CreateDateColumn({ type: 'timestamp with time zone', name: 'created_at' })
  createdAt!: Date;

  @UpdateDateColumn({ type: 'timestamp with time zone', name: 'updated_at' })
  updatedAt!: Date;

  get nextPaymentDate(): Date | null {
    if (this.frequency === MoneyRequestFrequency.ONE_TIME) {
      return null;
    }
    const startDate = moment(this.startDate).clone().startOf('day');
    const endDate = this.endDate ? moment(this.endDate).endOf('day') : null;
    const now = moment().startOf('day');

    if (endDate && moment().isAfter(endDate)) {
      return null;
    }

    const calculateNextDate = (unit: moment.unitOfTime.Diff) => {
      const diff = now.diff(startDate, unit);
      const nextDate = startDate.clone().add(diff, unit);

      const adjustedDate = nextDate.isSameOrAfter(now) ? nextDate : nextDate.add('1', unit);

      if (endDate && adjustedDate.isAfter(endDate)) {
        return null;
      }

      return adjustedDate.toDate();
    };

    switch (this.frequency) {
      case MoneyRequestFrequency.DAILY:
        return calculateNextDate('days');
      case MoneyRequestFrequency.WEEKLY:
        return calculateNextDate('weeks');
      case MoneyRequestFrequency.MONTHLY:
        return calculateNextDate('months');
      default:
        return null;
    }
  }
}
