import { Body, Controller, Get, HttpCode, HttpStatus, Param, Post, Query, UseGuards } from '@nestjs/common';
import { ApiBearerAuth, ApiTags } from '@nestjs/swagger';
import { Roles } from '~/auth/enums';
import { IJwtPayload } from '~/auth/interfaces';
import { AllowedRoles, AuthenticatedUser } from '~/common/decorators';
import { RolesGuard } from '~/common/guards';
import { ApiDataPageResponse, ApiDataResponse } from '~/core/decorators';
import { PageOptionsRequestDto } from '~/core/dtos';
import { CustomParseUUIDPipe } from '~/core/pipes';
import { ResponseFactory } from '~/core/utils';
import { createCategoryRequestDto, CreateGoalRequestDto, FundGoalRequestDto } from '../dtos/request';
import {
  CategoriesListResponseDto,
  CategoryResponseDto,
  GoalsStatsResponseDto,
  SavingGoalDetailsResponseDto,
} from '../dtos/response';
import { SavingGoalsService } from '../services';
import { CategoryService } from '../services/category.service';

@Controller('saving-goals')
@ApiTags('Saving Goals')
@UseGuards(RolesGuard)
@AllowedRoles(Roles.JUNIOR)
@ApiBearerAuth()
export class SavingGoalsController {
  constructor(
    private readonly savingGoalsService: SavingGoalsService,
    private readonly categoryService: CategoryService,
  ) {}

  @Post()
  @ApiDataResponse(SavingGoalDetailsResponseDto)
  async createGoal(@AuthenticatedUser() { sub }: IJwtPayload, @Body() body: CreateGoalRequestDto) {
    const goal = await this.savingGoalsService.createGoal(sub, body);

    return ResponseFactory.data(new SavingGoalDetailsResponseDto(goal));
  }

  @Get()
  @ApiDataPageResponse(SavingGoalDetailsResponseDto)
  async getGoals(@AuthenticatedUser() { sub }: IJwtPayload, @Query() pageOptions: PageOptionsRequestDto) {
    const [goals, itemCount] = await this.savingGoalsService.findGoals(sub, pageOptions);

    return ResponseFactory.dataPage(
      goals.map((goal) => new SavingGoalDetailsResponseDto(goal)),
      {
        page: pageOptions.page,
        size: pageOptions.size,
        itemCount,
      },
    );
  }

  @Get('stats')
  @ApiDataResponse(GoalsStatsResponseDto)
  async getStats(@AuthenticatedUser() { sub }: IJwtPayload) {
    const stats = await this.savingGoalsService.getStats(sub);

    return ResponseFactory.data(new GoalsStatsResponseDto(stats));
  }

  @Get('categories')
  @ApiDataResponse(CategoriesListResponseDto)
  async getCategories(@AuthenticatedUser() { sub }: IJwtPayload) {
    const categories = await this.categoryService.findCategories(sub);

    return ResponseFactory.data(new CategoriesListResponseDto(categories));
  }

  @Get(':goalId')
  @ApiDataResponse(SavingGoalDetailsResponseDto)
  async getGoal(@AuthenticatedUser() { sub }: IJwtPayload, @Param('goalId', CustomParseUUIDPipe) goalId: string) {
    const goal = await this.savingGoalsService.findGoalById(sub, goalId);

    return ResponseFactory.data(new SavingGoalDetailsResponseDto(goal));
  }

  @Post('categories')
  @ApiDataResponse(CategoryResponseDto)
  async createCategory(@AuthenticatedUser() { sub }: IJwtPayload, @Body() body: createCategoryRequestDto) {
    const category = await this.categoryService.createCustomCategory(sub, body);
    return ResponseFactory.data(new CategoryResponseDto(category));
  }

  @Post(':goalId/fund')
  @HttpCode(HttpStatus.NO_CONTENT)
  async fundGoal(
    @AuthenticatedUser() { sub }: IJwtPayload,
    @Param('goalId', CustomParseUUIDPipe) goalId: string,
    @Body() body: FundGoalRequestDto,
  ) {
    await this.savingGoalsService.fundGoal(sub, goalId, body);
  }
}
