import { BadRequestException, forwardRef, Inject, Injectable, Logger } from '@nestjs/common';
import { FindOptionsWhere } from 'typeorm';
import { CustomerNotificationSettings } from '~/customer/entities/customer-notification-settings.entity';
import { CustomerService } from '~/customer/services';
import { Guardian } from '~/guardian/entities/guradian.entity';
import { CreateUnverifiedUserRequestDto } from '../../auth/dtos/request';
import { Roles } from '../../auth/enums';
import { User } from '../entities';
import { UserRepository } from '../repositories';

@Injectable()
export class UserService {
  private readonly logger = new Logger(UserService.name);
  constructor(
    private readonly userRepository: UserRepository,
    @Inject(forwardRef(() => CustomerService)) private readonly customerService: CustomerService,
  ) {}

  findUser(where: FindOptionsWhere<User> | FindOptionsWhere<User>[]) {
    this.logger.log(`finding user with where clause ${JSON.stringify(where)}`);
    return this.userRepository.findOne(where);
  }

  async findUserOrThrow(where: FindOptionsWhere<User>) {
    this.logger.log(`Finding user with where clause ${JSON.stringify(where)}`);
    const user = await this.findUser(where);

    if (!user) {
      this.logger.error(`User with not found with where clause ${JSON.stringify(where)}`);
      throw new BadRequestException('USERS.NOT_FOUND');
    }

    this.logger.log(`User with where clause ${JSON.stringify(where)} found successfully`);
    return user;
  }

  async findOrCreateUser({ phoneNumber, countryCode }: CreateUnverifiedUserRequestDto) {
    this.logger.log(`Finding or creating user with phone number ${phoneNumber} and country code ${countryCode}`);
    const user = await this.userRepository.findOne({ phoneNumber });

    if (!user) {
      this.logger.log(`User with phone number ${phoneNumber} not found, creating new user`);
      return this.userRepository.createUnverifiedUser({ phoneNumber, countryCode, roles: [Roles.GUARDIAN] });
    }
    if (user && user.roles.includes(Roles.GUARDIAN) && user.isPasswordSet) {
      this.logger.error(`User with phone number ${phoneNumber} already exists`);
      throw new BadRequestException('USERS.PHONE_NUMBER_ALREADY_EXISTS');
    }

    if (user && user.roles.includes(Roles.JUNIOR)) {
      this.logger.error(`User with phone number ${phoneNumber} is an already registered junior`);
      throw new BadRequestException('USERS.JUNIOR_UPGRADE_NOT_SUPPORTED_YET');
      //TODO add role Guardian to the existing user and send OTP
    }

    this.logger.log(`User with phone number ${phoneNumber} and country code ${countryCode} found successfully`);
    return user;
  }

  async createUser(data: Partial<User>) {
    this.logger.log(`Creating user with data ${JSON.stringify(data)}`);
    const user = await this.userRepository.createUser(data);

    this.logger.log(`User with data ${JSON.stringify(data)} created successfully`);
    return user;
  }

  setEmail(userId: string, email: string) {
    this.logger.log(`Setting email ${email} for user ${userId}`);
    return this.userRepository.update(userId, { email });
  }

  setPasscode(userId: string, passcode: string, salt: string) {
    this.logger.log(`Setting passcode for user ${userId}`);
    return this.userRepository.update(userId, { password: passcode, salt, isProfileCompleted: true });
  }

  async verifyUserAndCreateCustomer(user: User) {
    this.logger.log(`Verifying user ${user.id} and creating customer`);
    await this.customerService.createCustomer(
      {
        guardian: Guardian.create({ id: user.id }),
        notificationSettings: new CustomerNotificationSettings(),
      },
      user,
    );

    this.logger.log(`User ${user.id} verified and customer created successfully`);
    return this.findUserOrThrow({ id: user.id });
  }
}
